/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 this example program is derived from the QuantLib test suite vpp.hpp & vpp.cpp

 compile:
g++ -I/home/spanderen/workspace/QuantLib -O3 -o vpp6 ql_vpp6.cpp -L/home/spanderen/workspace/QuantLib/ql/.libs -lQuantLib
*/

#include <ql/settings.hpp>
#include <ql/math/functional.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/math/generallinearleastsquares.hpp>
#include <ql/math/randomnumbers/sobolrsg.hpp>
#include <ql/math/statistics/generalstatistics.hpp>
#include <ql/models/marketmodels/browniangenerators/sobolbrowniangenerator.hpp>
#include <ql/instruments/vanillaswingoption.hpp>
#include <ql/methods/montecarlo/lsmbasissystem.hpp>
#include <ql/methods/montecarlo/multipathgenerator.hpp>
#include <ql/experimental/processes/klugeextouprocess.hpp>
#include <ql/experimental/processes/extouwithjumpsprocess.hpp>
#include <ql/experimental/processes/extendedornsteinuhlenbeckprocess.hpp>
#include <ql/methods/finitedifferences/meshers/uniform1dmesher.hpp>
#include <ql/methods/finitedifferences/meshers/fdmmeshercomposite.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>
#include <ql/methods/finitedifferences/utilities/fdminnervaluecalculator.hpp>
#include <ql/experimental/finitedifferences/fdmvppstepcondition.hpp>
#include <ql/experimental/finitedifferences/fdsimpleklugeextouvppengine.hpp>

#include <fstream>

using namespace QuantLib;

namespace {

    // model definition
    const Real beta         = 200;
    const Real eta          = 1.0/0.4;
    const Real lambda       = 4.0;
    const Real alpha        = 7.0;
    const Real volatility_x = 1.4;
    const Real kappa        = 4.45;
    const Real volatility_u = std::sqrt(1.3);
    const Real rho          = 0.7;
    const Rate irRate       = 0.00;

    // vpp definition
    const Size maturityInWeeks  = 1;
    const Real heatRate         = 2.5;
    const Real pMin             = 8;    // MWh
    const Real pMax             = 40;   // MWh
    const Real startUpFuel      = 20;   // MWh
    const Real startUpFixCost   = 300;  // EUR
    const Real carbonPrice      = 3.0;  // EUR per MWh heat (from gas)
    const Size tMinUp           = 6.0;  // hour
    const Size tMinDown         = 2.0;  // hour
    const Size nStates          = 2u*tMinUp + tMinDown;

    typedef PseudoRandom::rsg_type rsg_type;
    typedef MultiPathGenerator<rsg_type>::sample_type sample_type;
    typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;


    class PathGasPrice : public FdmInnerValueCalculator {
      public:
        typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;

        PathGasPrice(
            const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path,
            const boost::shared_ptr<Shape>& shape)
        : path_(path),
          shape_(shape) {}
        Real innerValue(const FdmLinearOpIterator&, Time t) {
            const Size i = (Size) (t*365*24);
            const Real f = std::lower_bound(shape_->begin(), shape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;

            return std::exp(path_[2][i] + f);
        }
        Real avgInnerValue(const FdmLinearOpIterator& iter, Time t) {
            return innerValue(iter, t);
        }
      private:
        const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path_;
        const boost::shared_ptr<Shape> shape_;
    };

    class PathSparkSpreadPrice : public FdmInnerValueCalculator {
      public:
        typedef FdSimpleKlugeExtOUVPPEngine::Shape Shape;

        PathSparkSpreadPrice(
            Real heatRate,
            const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path,
            const boost::shared_ptr<Shape>& gasShape,
            const boost::shared_ptr<Shape>& powerShape)
        : heatRate_(heatRate),
          path_(path),
          gasShape_(gasShape),
          powerShape_(powerShape) {}

        Real innerValue(const FdmLinearOpIterator&, Time t) {
            const Size i = (Size) (t*365*24);
            const Real f = std::lower_bound(
                powerShape_->begin(), powerShape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;
            const Real g = std::lower_bound(
                gasShape_->begin(),gasShape_->end(),
                std::pair<Time, Real>(t-std::sqrt(QL_EPSILON), 0.0))->second;
            
            return std::exp(f + path_[0][i]+path_[1][i])
                    - heatRate_*std::exp(g + path_[2][i]);
        }
        Real avgInnerValue(const FdmLinearOpIterator& iter, Time t) {
            return innerValue(iter, t);
        }
      private:
        const Real heatRate_;
        const MultiPathGenerator<PseudoRandom>::sample_type::value_type& path_;
        const boost::shared_ptr<Shape> gasShape_;
        const boost::shared_ptr<Shape> powerShape_;
    };

    class SobolQRsg {
    public:
        typedef SobolRsg::sample_type sample_type;

        explicit SobolQRsg(Size dimensionality, unsigned long seed = 1234u)
        : rsg_(dimensionality, 1234, SobolRsg::JoeKuoD7),
          seq_(std::vector<Real>(dimensionality), 1.0),
          offsets_(dimensionality) {
            MersenneTwisterUniformRng rng(seed);
            for (Size i=0; i < dimensionality; ++i) {
                offsets_[i] = rng.nextReal();
            }
        }

        const SobolQRsg::sample_type& nextSequence() const {
            seq_ = rsg_.nextSequence();
            for (Size i=0; i < seq_.value.size(); ++i) {
                seq_.value[i]+=offsets_[i];
                if (seq_.value[i] >= 1.0) {
                    seq_.value[i] -=1.0;
                }
            }
            return seq_;
        }

        const sample_type& lastSequence() const { return seq_; }
        Size dimension() const { return rsg_.dimension(); }
    private:
        SobolRsg rsg_;
        mutable sample_type seq_;
        std::vector<Real> offsets_;
    };

    class SobolBrownianSequenceGenator {
      public:
        typedef Sample<std::vector<Real> > sample_type;

        SobolBrownianSequenceGenator(Size factors, Size steps, 
                                     unsigned long seed = 1234u)
        : dim_(factors*steps),
          factors_(factors),
          steps_(steps),
          qrsg_(SobolQRsg(dim_, seed)), 
          gen_(factors, steps, SobolBrownianGenerator::Diagonal),
          seq_(std::vector<Real>(dim_), 1.0),
          variates_(dim_, std::vector<Real>(1u)) {
        }
              
        const sample_type& nextSequence() const {
            const SobolQRsg::sample_type& sample = qrsg_.nextSequence();
            for (Size k=0; k < dim_; ++k) {
                variates_[k].front() = sample.value[k];
            }
            const std::vector<std::vector<Real> > bb = gen_.transform(variates_);
            for (Size i=0; i < factors_; ++i) {
                for (Size j=0; j < steps_; ++j) {
                    seq_.value[j*factors_+i] = bb[i][j];
                }
            }

            return seq_;
        }
        const sample_type& lastSequence() const {
            return seq_;
        }
        Size dimension() const { return dim_; }

    private:
        const Size dim_;
        const Size factors_;
        const Size steps_;
        const InverseCumulativeRsg<SobolQRsg, InverseCumulativeNormal> qrsg_;

        mutable SobolBrownianGenerator gen_;
        mutable sample_type seq_;
        mutable std::vector<std::vector<Real> > variates_;
    };
}

int main(int argc, char *argv[]) {

    const DayCounter dc = ActualActual();
    const Date today = Date(28, Sep, 2010);
    Settings::instance().evaluationDate() = today;

    const Size maturityInDays  = 7*maturityInWeeks;
    const Size maturityInHours = 24*maturityInDays;
    Date maturityDate = today + Period(maturityInWeeks, Weeks);

    boost::shared_ptr<Shape> gasShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> powerShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> gasZeroVolShape(new Shape(maturityInHours));
    boost::shared_ptr<Shape> powerZeroVolShape(new Shape(maturityInHours));

    const boost::shared_ptr<SwingExercise> exercise(
        new SwingExercise(today, today+maturityInDays-1, 3600u));
    const std::vector<Time> exerciseTimes = exercise->exerciseTimes(dc, today);

    // read shape from file
    // format is Date, 0h price, 1h price, 2h price, ... 23h price
    // example:4-Oct-2010 36.73 32.09 27.32 23.22 25.71 35.60 47.32 ...
    std::ifstream eex("EEX_2010-2013.txt");
    QL_REQUIRE(eex.is_open(), "can't find price file EEX_2010-2013.txt");
    
    std::ifstream ttf("TTF_2010-2013.txt");
    QL_REQUIRE(ttf.is_open(), "can't find price file TTF_2010-2013.txt");

    Real p;
    for (Size i=0; i < maturityInDays; ++i) {
        std::string str;
        eex >> str;
        ttf >> str;
        
        for (Size j=0; j < 24; ++j) {
            const Size ij = i*24+j;
            const Time t = exerciseTimes[ij]+1/(365*24.); 
            
            eex >> p;
            (*powerShape)[ij] = Shape::value_type(t, 
                 std::log(p)-square<Real>()(volatility_x)
                             /(4*alpha)*(1-std::exp(-2*alpha*t))
                 -lambda/beta*std::log((eta-std::exp(-beta*t))/(eta-1.0))); 
            (*powerZeroVolShape)[ij] = Shape::value_type(t, std::log(p));

            ttf >> p;
            (*gasShape)[ij] = Shape::value_type(t, 
                 std::log(p)-square<Real>()(volatility_u)
                             /(4*kappa)*(1-std::exp(-2*kappa*t)));
            (*gasZeroVolShape)[ij] = Shape::value_type(t, std::log(p));
        }
    }
    eex.close();
    ttf.close();
 
    const boost::shared_ptr<FdmMesher> oneDimMesher(
        new FdmMesherComposite(boost::shared_ptr<FdmLinearOpLayout>(
            new FdmLinearOpLayout(std::vector<Size>(1, nStates))),
                std::vector<boost::shared_ptr<Fdm1dMesher> >(1,
                    boost::shared_ptr<Fdm1dMesher>(
                        new Uniform1dMesher(0.0, 1.0, nStates)))));

    const TimeGrid grid(dc.yearFraction(today, exercise->lastDate()+1),
                        exercise->dates().size());

    Array x0(2);
    x0[0] = 0.0; x0[1] = 0.0;

    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> ouProcess(
        new ExtendedOrnsteinUhlenbeckProcess(alpha, volatility_x, x0[0],
                                             constant<Real, Real>(x0[0])));
    boost::shared_ptr<ExtOUWithJumpsProcess> lnPowerProcess(
           new ExtOUWithJumpsProcess(ouProcess, x0[1], beta, lambda, eta));
    
    Real u=0.0;
    boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> lnGasProcess(
        new ExtendedOrnsteinUhlenbeckProcess(kappa, volatility_u, u,
                                             constant<Real, Real>(u)));
    
    const boost::shared_ptr<KlugeExtOUProcess> klugeOUProcess(
        new KlugeExtOUProcess(rho, lnPowerProcess, lnGasProcess));

    rsg_type rsg = PseudoRandom::make_sequence_generator(
        klugeOUProcess->factors()*(grid.size()-1), 1234ul);

    const boost::shared_ptr<MultiPathGenerator<rsg_type> > generator(
        new MultiPathGenerator<rsg_type>(klugeOUProcess, grid, rsg, false));

    // Test: Longstaff Schwartz least square Monte-Carlo
    const Size nCalibrationTrails = 4096u;
    const Size nTrails = 2048u;

    GeneralStatistics npv;
    std::vector<sample_type> calibrationPaths;
    std::vector<boost::shared_ptr<FdmVPPStepCondition> > stepConditions;
    std::vector<boost::shared_ptr<FdmInnerValueCalculator> > sparkSpreads;

    sparkSpreads.reserve(nCalibrationTrails);
    stepConditions.reserve(nCalibrationTrails);
    calibrationPaths.reserve(nCalibrationTrails);

    for (Size i=0; i < nCalibrationTrails; ++i) {
        calibrationPaths.push_back(generator->next());

        sparkSpreads.push_back(boost::shared_ptr<FdmInnerValueCalculator>(
            new PathSparkSpreadPrice(heatRate, calibrationPaths.back().value,
                                     gasShape, powerShape)));

        stepConditions.push_back(boost::shared_ptr<FdmVPPStepCondition>(
            new FdmVPPStepCondition(
                heatRate, pMin, pMax, tMinUp, tMinDown,
                startUpFuel, startUpFixCost, carbonPrice, 0u, oneDimMesher,
                boost::shared_ptr<FdmInnerValueCalculator>(
                    new PathGasPrice(calibrationPaths.back().value, gasShape)),
                sparkSpreads.back())));
    }

    const FdmLinearOpIterator iter = oneDimMesher->layout()->begin();

    // prices of all calibration paths for all states
    std::vector<Array> prices(nCalibrationTrails, Array(nStates, 0.0));

    // regression coefficients for all states and all exercise dates
    std::vector<std::vector<Array> > coeff(
        nStates, std::vector<Array>(exercise->dates().size(), Array()));

    // regression functions

    const Size dim = 2u;
    std::vector<boost::function1<Real, Array> > v(
        LsmBasisSystem::multiPathBasisSystem(dim, 3u, LsmBasisSystem::Monomial));

    for (Size i=exercise->dates().size(); i > 0u; --i) {
        const Time t = grid.at(i);

        std::vector<Array> x(nCalibrationTrails, Array(dim));

        for (Size j=0; j < nCalibrationTrails; ++j) {
            x[j][0] = sparkSpreads[j]->innerValue(iter, t);
            x[j][1] = calibrationPaths[j].value[1][i]; 
        }

        for (Size k=0; k < nStates; ++k) {
            std::vector<Real> y(nCalibrationTrails);

            for (Size j=0; j < nCalibrationTrails; ++j) {
                y[j] = prices[j][k];
            }
            coeff[k][i-1] = GeneralLinearLeastSquares(x, y, v).coefficients();

            for (Size j=0; j < nCalibrationTrails; ++j) {
                prices[j][k] = 0.0;
                for (Size l=0; l < v.size(); ++l) {
                    prices[j][k] += coeff[k][i-1][l]*v[l](x[j]);
                }
            }
        }

        for (Size j=0; j < nCalibrationTrails; ++j) {
            stepConditions[j]->applyTo(prices[j], grid.at(i));
        }
    }


    const Size steps   = grid.size()-1;
    const Size factors = klugeOUProcess->factors();

    npv.reset();
    MultiPathGenerator<SobolBrownianSequenceGenator> sobolGen(
        klugeOUProcess, grid, SobolBrownianSequenceGenator(factors, steps), false);

    for (Size i=0; i < nTrails; ++i) {
        const sample_type& path = sobolGen.next();

        const boost::shared_ptr<FdmInnerValueCalculator> gasPrices(
            new PathGasPrice(path.value, gasShape));

        const boost::shared_ptr<FdmInnerValueCalculator> sparkSpreads(
            new PathSparkSpreadPrice(heatRate, path.value,
                                     gasShape, powerShape));

        FdmVPPStepCondition stepCondition(
            heatRate, pMin, pMax, tMinUp, tMinDown,
            startUpFuel, startUpFixCost, carbonPrice, 0u, oneDimMesher,
            gasPrices, sparkSpreads);

        Array x(dim), state(nStates, 0.0), contState(nStates, 0.0);
        for (Size j=exercise->dates().size(); j > 0u; --j) {
            const Time t = grid.at(j);
            const Real gasPrice = gasPrices->innerValue(iter, t);
            const Real sparkSpread = sparkSpreads->innerValue(iter, t);
            const Real startUpCost
                    = startUpFixCost + (gasPrice + carbonPrice)*startUpFuel;

            x[0] = sparkSpread;
            x[1] = path.value[1][j];
            for (Size k=0; k < nStates; ++k) {
                contState[k] = 0.0;
                for (Size l=0; l < v.size(); ++l) {
                    contState[k] += coeff[k][j-1][l]*v[l](x);
                }
            }

            const Real pMinFlow = pMin*(sparkSpread - heatRate*carbonPrice);
            const Real pMaxFlow = pMax*(sparkSpread - heatRate*carbonPrice);

            // rollback continuation states and the path states
            for (Size i=0; i < 2*tMinUp; ++i) {
                if (i < tMinUp) {
                    state[i]    += pMinFlow;
                    contState[i]+= pMinFlow;
                }
                else {
                    state[i]    += pMaxFlow;
                    contState[i]+= pMaxFlow;
                }
            }

            // dynamic programming using the continuation values
            Array retVal(nStates);
            for (Size i=0; i < tMinUp-1; ++i) {
                retVal[i] = retVal[tMinUp + i]
                          = (contState[i+1] > contState[tMinUp + i+1])?
                                          state[i+1] : state[tMinUp + i+1];
            }

            if (contState[2*tMinUp] >=
                std::max(contState[tMinUp-1], contState[2*tMinUp-1])) {
                retVal[tMinUp-1] = retVal[2*tMinUp-1] = state[2*tMinUp];
            }
            else if (contState[tMinUp-1] >= contState[2*tMinUp-1]) {
                retVal[tMinUp-1] = retVal[2*tMinUp-1] = state[tMinUp-1];
            }
            else {
                retVal[tMinUp-1] = retVal[2*tMinUp-1] = state[2*tMinUp-1];
            }

            for (Size i=0; i < tMinDown-1; ++i) {
                retVal[2*tMinUp + i] = state[2*tMinUp + i+1];
            }

            if (contState.back() >=
                std::max(contState.front(), contState[tMinUp]) - startUpCost) {
                retVal.back() = state.back();
            }
            else if (contState.front() >  contState[tMinUp]) {
                retVal.back() = state.front()-startUpCost;
            }
            else {
                retVal.back() = state[tMinUp]-startUpCost;
            }
            state = retVal;
        }
        npv.add(state.back());
    }

    printf("%f \n",npv.mean());

    return 0;
}
